"""
Model2Mass Subpackage
=====================

This subpackage provides tools for constructing and analyzing
particle-physics models based on mass matrices.

It is designed to interface seamlessly with the `PyDiscrete`
subpackage, which supplies finite group–theory utilities
via the GAP system.

"""

# Import main user-facing API from the module
from .Model2Mass import Model2Mass, make_latex_tableA4, make_latex_tableT19

__all__ = ["Model2Mass", "make_latex_tableA4", "make_latex_tableT19"]
__version__ = "0.2.0"



import numpy as np
import sympy as sp
import os
import itertools
import subprocess
import ast
import re
from pathlib import Path
import importlib.resources as ir
import pickle

class Group():
 
    """
     This class represents a discrete group obtained by GAP. The functions in this class can compute group theory objects such as CG coefficients or tensor products. You can also define fields which transform under the different irreducible representations of the group. Basic information from the group is obtained through GAP but this is stored in this class too.
     
     Attributes
     ----------
     
       - name: str
           This is the name of the group such that GAP can read it. It is always of the form "SmallGroup(gapID1, gapID2)". Here gapID1 and gapID2 are integers used only during initialization.
        
       - GAPoutput: str
           Output obtained from calling GAP. It shows the basic information such as the character table, matrix representations and size of conjugacy classes.
        
       - GAPerror: str
           Error printed by GAP in case the group was not called.

       -  GAP_run : boolean
           Boolean that tells whether GAP was used or the group was taken from the saved_groups.pkl file.
    
       - structure_description: str
           Name given by GAP.
       
       - size: int
           Size of the group.
       
       - rep_matrices: list
           list of the irreducbible representation matrices of the group. The matrices are numpy arrays.
       
       - dimensions_of_reps: list
           Contains the dimensions of the irreducible representations of the group.
    
       - number_of_reps: int
           number of irreducible representaations of the group.
    
       - size_conjugacy_classes : list
           Sizes of the conjugacy classes of the group.
    
       - character_table: numpy array
          Character tabel of the group.
      
       - unitary_representations: bool
          If True, the group has all unitary representations. This is computed by PyDiscrete when the  Group class is initialized.
        
       - cg_coeff_calculated: numpy array
          Contains boolean values of the CG coeffficients calculated for the different products of the group. It is initialized as a matrix of zeroes.
        
       - cg_coeff: numpy array
          Contains the float CG coefficients for the different products of the group.
        
       - kronecker_product_calculated: numpy array
          Contains boolean values of whether the kronecker products (multiplicities of the sum of irreducible representations obtained from the product of two irreps) have been previously calculated.
        
       - kronecker_product: numpy array
          Contains the kronecker products (multiplicities of the irreducible representations obtained from the product of two irreps).

       
    
    
    Methods:
    --------
    
    
       - isUnitary():
          Checks if a group is unitary
        
       - isUnitaryRep():
          Checks if a irrep is unitary
        
       - getCGC(p,q):
          Returns a numpy array with the CG coefficients for the product  of the p-th irrep times the q-th irrep. The ordering of p,q is based on the order of the character table
        
       - getCGCcc(p,q):
          Returns the complex conjugate transpose  of the array given by getCGC, if the group is unitary. Otherwise, it gives the inverse of the array given by getCGC
        
       - getKroneckerProduct(p,q):
          Returns a numpy array of the multiplicity of the irreducible representations obtained in the decomposition of the product of the pth irrep and the qth irrep
        
       - getMultiplicity(p,q,r):
          Returns an integer which corresponds to the multiplicity of the r-th irrep in the product of the p-th and q-th irreps.
        
       - getRepVector(pRep, label_p):
          Returns a list with vectors of sympy symbols using label_p as a lable. This list corresponds to a field transforming in the pRep-th irrep.
        
       - multiply(list_fields):
          Returns a list that contains the irrep decomposition of the product of the fields in the list_fields
        
       - extractOneRep(vec_rep, idx_p):
          Returns a vector with only the idx_p-th irrep in the list of irreps in vec_rep

    """

   

    def __init__(self, gapID1, gapID2, gap_path, compute_ALL_CGC = True):
        """
        
        Input:
        ----------
        
          - gapID1: integer
             First number of the GAP ID
        
          - gapID2: integer
             Second number of the GAP ID
        
          - gap_path: str
             Path for the location of GAP

          - compute_ALL_CGC: boolean, optional
             If True, it computes all CGC coefficients when initialized
        
        
        """
        self.name = "SmallGroup(" + str(gapID1) + "," + str(gapID2) + ")"

        
        with ir.as_file(
                ir.files("FlavorBuilder.PyDiscrete") / "saved_groups.pkl"
        ) as saved_groups_pickle_path:
            with open(saved_groups_pickle_path, "rb") as f:
                self.saved_groups_dict = pickle.load(f)

        if (gapID1, gapID2) in self.saved_groups_dict:
            self.GAPoutput = self.saved_groups_dict[(gapID1, gapID2)]
            self.GAP_run = False

        else:
            # Path for .gap file
            gap_exe = Path(gap_path).expanduser().resolve()
            if not gap_exe.exists() or not os.access(gap_exe, os.X_OK):
                raise RuntimeError(
                    f"Invalid GAP executable path: '{gap_exe}'. "
                    "Provide the full path to the GAP binary and ensure it is executable."
                    )

            #Prepare gap script to extract group info
            with ir.as_file(
                    ir.files("FlavorBuilder.PyDiscrete") / "GAP-Script" / "extractGroupDataPy.gap"
            ) as gap_script_path:
                    # Make a GAP-safe literal path (forward slashes; escape quotes)
                    gap_script_literal = '"' + gap_script_path.as_posix().replace('"', r'\"') + '"'


                    # 3) Build batch code exactly like your original logic
                    gap_code = (
                        f"Read({gap_script_literal});; "
                        f"MBgetGroupInfo({self.name});; "
                        f"quit;;"
                        )


            # Run GAP as a subprocess
            result = subprocess.run(
                [str(gap_exe), "-q", "-b"],
                input=gap_code.encode("utf-8"),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=False,
            )

            self.GAPoutput = result.stdout.decode("utf-8", "replace").strip()
            self.GAPerror = result.stderr.decode("utf-8", "replace").strip()
        
            #Prints the GAP error if the group properties were not sucesfully obtained from GAP
            if result.returncode != 0:
                raise RuntimeError(
                    f"GAP execution failed for {self.name} (exit code {result.returncode}).\n"
                    f"STDERR:\n{self.GAPerror}"
                )

            if self.GAPerror:
                print(f"[Warning] GAP reported messages for {self.name}:\n{self.GAPerror}\n")
            else:
                self.saved_groups_dict[(gapID1, gapID2)] = self.GAPoutput
        

        #Looks for the information in the GAPoutput
        match_StructureDescription = re.search(r"StructureDescription:\s*(.*?)\s*Size:",self.GAPoutput,re.DOTALL)
        match_Size = re.search(r"Size:\s*(.*?)\s*DimensionsOfReps:", self.GAPoutput, re.DOTALL)
        match_RepMatrices = re.search(r"RepMatrices:\s*(.*)", self.GAPoutput, re.DOTALL)
        match_DimensionsOfReps = re.search(r"DimensionsOfReps:\s*(.*?)\s*SizesConjugacyClasses:", self.GAPoutput, re.DOTALL)
        match_SizeConjugacyClasses = re.search(r"SizesConjugacyClasses:\s*(.*?)\s*CharacterTable:", self.GAPoutput, re.DOTALL)
        match_CharacterTable = re.search(r"CharacterTable:\s*(.*?)\s*RepMatrices:", self.GAPoutput, re.DOTALL)


        #Replace GAP output for actual numerical values
            
        # First replace E(n)^k (must be done before the plain E(n) case)
        s_RepMatrices = re.sub(r"E\((\d+)\)\^(\d+)", self._replace_E_power, match_RepMatrices.group(1))
        s_CharacterTable = re.sub(r"E\((\d+)\)\^(\d+)", self._replace_E_power, match_CharacterTable.group(1))

        # Then replace plain E(n)
        s_RepMatrices = re.sub(r"E\((\d+)\)", self._replace_E_plain, s_RepMatrices)
        s_CharacterTable = re.sub(r"E\((\d+)\)", self._replace_E_plain, s_CharacterTable)



        # Use eval, but restrict to only allow numpy functions
        safe_globals = {"np": np}
        nested_RepMatrices = eval(s_RepMatrices, safe_globals)
        nested_CharacterTable = eval(s_CharacterTable, safe_globals)



        # Convert nested list into list of numpy arrays
        final_RepMatrices = [[np.array(entry) for entry in row] for row in nested_RepMatrices]
        final_CharacterTable = np.array([[entry for entry in row] for row in nested_CharacterTable])

        
        
        #Define the attributes of the group class
        self.structure_description = match_StructureDescription.group(1).replace('"', '')
        self.size = int(match_Size.group(1))
        self.rep_matrices = final_RepMatrices
        self.dimensions_of_reps = ast.literal_eval(match_DimensionsOfReps.group(1))
        self.number_of_reps = len(ast.literal_eval(match_DimensionsOfReps.group(1)))
        self.size_conjugacy_classes = ast.literal_eval(match_SizeConjugacyClasses.group(1))
        self.character_table = final_CharacterTable
        self.unitary_representations = self.isUnitary()
        self.cg_coeff_calculated = np.zeros((self.number_of_reps, self.number_of_reps), dtype=bool)
        self.cg_coeff = np.zeros((self.number_of_reps, self.number_of_reps), dtype=object)
        self.kronecker_product_calculated = np.zeros((self.number_of_reps, self.number_of_reps), dtype=bool)
        self.kronecker_product = np.zeros((self.number_of_reps, self.number_of_reps), dtype=object)
        self.NumberOfSinglets = np.count_nonzero(np.array(self.dimensions_of_reps) == 1)

        # If compute_ALL_CGC = True, then all CG coefficients are computed when the group is initialized
        if compute_ALL_CGC:
            for li in range(1, self.number_of_reps + 1):
                for lj in range(1, self.number_of_reps + 1):
                    self.getCGC(li, lj)
                
                
                
    def isUnitary(self):
    
        """
        Checks if the group is unitary.
        
        Output:
        -------
        
          - unitaryQ:Boolean.
              If true, the group is unitary.
        """
        
        num_reps = self.number_of_reps
        unitaryQ = True
        for li in range(num_reps):
            lli = li +1
            if not self.isUnitaryRep(lli):
                unitaryQ = False
                return unitaryQ
        return unitaryQ
    
    def isUnitaryRep(self,rep_idx_P):
    
        """
        Checks whether the specified irrep is unitary. The irreducible representations are labeled according to their order in the character table obtained from GAP. The function tests if the `rep_idx_P`-th representation is unitarty.

        
        Input:
        ----------
          - rep_idx_P: integer
               Index of the irrep (1-based) whose unitarity is to be checked. The ordering is based in the character table obtained from GAP.
           
        Output:
        -------
           - unitaryrepQ: Boolean.
               If true, the rep_idx_P-th irrep is unitary. False otherwise.
        
        """
        
        if not (1 <= rep_idx_P <= len(self.dimensions_of_reps)):
            raise ValueError(
            f"Invalid irrep index: rep_idx_P = {rep_idx_P}. It must be between 1 and {len(self.dimensions_of_reps)} (inclusive)."
            )
        
        rep_idx_P = rep_idx_P -1
        #Checks if all matrices in the irrep rep_idx_P satisfy U^{\dagger} U = I
        rep_matrices = self.rep_matrices[rep_idx_P]
        products = [np.conj(U.T) @ U for U in rep_matrices]
        simplified = [np.round(P, decimals=6) for P in products]
        unique_matrices = {tuple(map(tuple, P)) for P in simplified}
        if len(unique_matrices) != 1:
            unitary_representationsREP = False
        
        identity = (np.eye(rep_matrices[0].shape[0]))
        
        unitaryrepQ = np.allclose(list(unique_matrices)[0], identity)

        return unitaryrepQ
    
    
    def _replace_E_power(self, some_gap_str):
    
        """
        This is a helper function. It replaces  E(n)^k for np.exp(2j * np.pi * {k} / {n}) in the GAP output.
        
        Input:
        -----------
        
           - some_gap_str: str
               String obtained from GAP output.
            
        
        Output:
        -------
        
            - repstr: str.
               A string where E(n)^k gets replaced by np.exp(2j * np.pi * {k} / {n}).
        """
        
        n, k = some_gap_str.group(1), some_gap_str.group(2)
        
        repstr = f"np.exp(2j * np.pi * {k} / {n})"
        return repstr

    def _replace_E_plain(self, some_gap_str):

        """
        This is a helper function. It replaces  E(n) for np.exp(2j * np.pi  / {n}) in the GAP output.
        
        Input:
        -----------
        
            - some_gap_str: str
               String obtained from GAP output.
            
        
        Output:
        -------
        
            - repstr: str.
               A string where E(n) gets replaced by np.exp(2j * np.pi  / {n}).
        """
        
        n = some_gap_str.group(1)
        repstr = f"np.exp(2j * np.pi / {n})"
        return repstr

    def _mIndex(self, dim_q, li, lj):
        """
        Helper function to calculate the CG coefficients.
        
        Parmeters
        ---------
        
            - dim_q: int.
                Dimension of second irrep q. The irreducible representations are labeled according to their order in the character table obtained from GAP.
           
            - li: int.
                Index for the first irrep p. Must be between 0 and the sp (dimension of p irrep).
           
            - lj: int.
                Index for the second irrep q (whose dimension it not a parameter in this function). Must be between 0 and the dim_q (dimension of q irrep).
           
        Output:
        -------
        
            - mindex: int.
                Integer given by (li-1) * dim_q  + lj.
    
        
        """
        
        mindex = (li - 1) * dim_q + lj
        return mindex

    def _pickTuple(self, dim_r, dim_p, dim_q, Amat, plstart=0):
    
        """
        Checks if one of the entries in the diagonal of the matrix Amat is non-zero. This is a helper function to construct the CG coeffcients.
        
        
        Input:
        ----------
        
            - dim_r: int
                Dimension of r irrep. The irreducible representations are labeled according to their order in the character table obtained from GAP.
            - dim_p: int
                Dimension of p irrep. The irreducible representations are labeled according to their order in the character table obtained from GAP.
            - dim_q: int
                Dimension of q irrep. The irreducible representations are labeled according to their order in the character table obtained from GAP.
            - Amat: numpy array
                Numpy array with matrices obtained from the helper function _A() from which can be obtained all the CG coefficients.
            - plstart: int, optional
                Index for where to start looking in the diagonal. Default is zero. If zero, then it looks over all the diagonal.
           
           
        Output:
        -------
            - ptuple: list of integers.
                List of integers if one of the entries in the diagonal of the matrix Amat is non-zero. Otherwise it returns none.
           
        """
        lstatus = True
        ll = plstart

        while lstatus and ll < dim_r:
            ll += 1
            li = 0
            
            while lstatus and li < dim_p:
                li += 1
                lj = 0
                
                while lstatus and lj < dim_q:
                    lj += 1
                    if Amat[ll-1][ll-1][self._mIndex(dim_q, li, lj) - 1, self._mIndex(dim_q, li, lj) - 1] != 0:
                        lstatus = False

        if lstatus:
            ptuple = None
            return ptuple
        else:
            ptuple = (li, lj, ll)
            return ptuple

    def _VectorZeroQ(self, vec):
    
        """
        Checks if a vector is non-zero
        
        Input:
        -----------
        
            - vec: numpy array
                The vector to be tested if it is non-zero.
            
            
        Output:
        -------
        
            - zeroQ: Boolean.
                It is True if vec is a non-zero vector.
        
        """
        lLen = len(vec)
        
        if lLen == 0:
            zeroQ = False
            return zeroQ
        
        if np.array_equal(vec, np.zeros(lLen)):
            zeroQ = True
            return zeroQ
        else:
            zeroQ = False
            return zeroQ

    def _getC(self, dim_r, dim_p, dim_q, l, Amat, plstart=0):
    
        """
        This is a helper function to calculate the CG coefficients.
        
        Input:
        ----------
        
            - dim_r: int
                 Dimension of the r irrep. The irreducible representations are labeled according to their order in the character table obtained from GAP.
            - dim_p: int
                 Dimension of the p irrep. The irreducible representations are labeled according to their order in the character table obtained from GAP.
            - dim_q: int
                Dimension of the q irrep. The irreducible representations are labeled according to their order in the character table obtained from GAP.
            - l: int
                Number between 0 and dim_r.
            - Amat: numpy array
                Numpy array with matrices obtained from the helper function _A() from which can be obtained all the CG coefficients.
            - plstart: int, optional
                Index for where to start looking for non-zero entries in the diagonal of the matrix Amat. Default is zero. If zero, then it looks over all the diagonal.
           
           
        Output:
        -------
        
            - getCarray: numpy array.
                A numpy array with specific elements from the matrix Amat. This will help obtain the CG coefficients to get the q-th irrep from the product of the r-th and p-th irrep.
        """
        
        li, lj, lk = self._pickTuple(dim_r, dim_p, dim_q, Amat, plstart)
        
        result = []
        for m in range(1, dim_p + 1):
            for n in range(1, dim_q + 1):
                result.append(
                    Amat[l-1][lk-1][self._mIndex(dim_q, m, n) - 1, self._mIndex(dim_q, li, lj) - 1]
                )
                
        getCarray = np.array(result).flatten()
        
        return getCarray


    def _getCNormalized(self, dim_r, dim_p, dim_q, Amat):
    
        """
        This is a helper function to calculate the CG coefficients when the group is unitary.
        
        Input:
        ----------
        
            - dim_r: int
                Dimension of the r irrep. The irreducible representations are labeled according to their order in the character table obtained from GAP.
            - dim_p: int
                Dimension of the p irrep. The irreducible representations are labeled according to their order in the character table obtained from GAP.
            - dim_q: int
                Dimension of the q irrep. The irreducible representations are labeled according to their order in the character table obtained from GAP.
            - Amat: numpy array
                Numpy array with matrices obtained from the helper function _A() from which can be obtained all the CG coefficients
           
           
        Output:
        -------
        
            - getCarraynom: numpy array.
                A numpy array with specific elements from the matrix Amat. This will help obtain the CG coefficients to get the q-th irrep from the product of the r-th and p-th irrep.
        """
    
        
        li, lj, lk = self._pickTuple(dim_r, dim_p, dim_q, Amat)
        result = []
        for ll in range(1, dim_r + 1):
            flattened_list = []
            for m in range(1, dim_p + 1):
                for n in range(1, dim_q + 1):
                    flattened_list.append(
                        Amat[ll-1][lk-1][self._mIndex(dim_q, m, n) - 1, self._mIndex(dim_q, li, lj) - 1]
                    )
            result.append(np.array(flattened_list))
        
        normalization_factor = np.sqrt(
            Amat[lk-1][lk-1][self._mIndex(dim_q, li, lj) - 1, self._mIndex(dim_q, li, lj) - 1]
        )
        
        getCarraynom = np.array(result) / normalization_factor
        
        return getCarraynom


    def _A(self, idx_r, idx_p, idx_q, pl, pk):
    
        """
        Calculates the matrix A used to obtain the CG coefficients. This is a helper function.
        
        Input:
        ----------
        
            - idx_r: int
                Index (1-based) of the r irrep according to character table. The ordering is based in the character table obtained from GAP.
            - idx_p: int
                Index (1-based) of the p irrep according to character table. The ordering is based in the character table obtained from GAP.
            - idx_q: int
                Index (1-based) of the q irrep according to character table. The ordering is based in the character table obtained from GAP.
            - pl: int
               index between 0 and the dimension of the r irrep.
            - pk: int
                index between 0a nd the dimension of the k irrep.
        
        Output:
        -------
        
            - matrixA: a numpy array.
                A numpy array with the matrix A which will be usted to get the CG coefficients.
        """
        def compute_sum(T, m, i, n, j): #Computes the summand for eq. (9) in Van den Broek, P. M., & Cornwell, J. F. (1978)
            return (
                self.rep_matrices[idx_p-1][T-1][m-1][i-1] *
                self.rep_matrices[idx_q-1][T-1][n-1][j-1] *
                np.conjugate(self.rep_matrices[idx_r-1][T-1][pl-1][pk-1])
                   )

        def compute_inner(m, n, i, j): #Computes the sum of eq. (9) in Van den Broek, P. M., & Cornwell, J. F. (1978)
            return sum(
                compute_sum(T, m, i, n, j)
                for T in range(1, self.size + 1)
            ) * self.dimensions_of_reps[idx_r-1] / self.size


        result = np.array([
                [
                    np.array([
                        [
                            compute_inner(m, n, i, j)
                            for j in range(1, self.dimensions_of_reps[idx_q-1] + 1)
                        ]
                        for i in range(1, self.dimensions_of_reps[idx_p-1] + 1)
                    ]).flatten()
                    for n in range(1, self.dimensions_of_reps[idx_q-1] + 1)
                ]
                for m in range(1, self.dimensions_of_reps[idx_p-1] + 1)
            ])
        matrixA = np.array([item for sublist in result for item in sublist])
        return matrixA


    def _generateKroneckerProduct(self, idx_p, idx_q):
    
        """
        Calculates a numpy array with the multiplicity of the irreducible representations obtained in the decomposition of the product of the p irrep and the q irrep.
        
        Input:
        ----------
        
            - idx_p: int
                Index (1-based) of the p irrep according to character table. The ordering is based in the character table obtained from GAP.
            - idx_q: int
                Index (1-based) of the q irrep according to character table. The ordering is based in the character table obtained from GAP.
          
        Output:
        -------
        
            - kroneckerarray: numpy array.
                A numpy array with the multiplicities of all the irreps in the product of the irreps p x q.
        
        """
        
        # Extract the character table
        character_table = self.character_table
        size_conjugacy_classes = np.array(self.size_conjugacy_classes)
        group_size = self.size

        #Calculate multiplicites from character formula
        product = character_table[ idx_p-1,:] * character_table[idx_q-1,:] * size_conjugacy_classes
        result = np.dot(character_table, np.conjugate(product)) / group_size
        complex_array = result
        
        
        #Round up to threshold
        threshold = 1e-5

        # Chop small real parts
        real_part = np.real(complex_array)
        real_part[np.abs(real_part) < threshold] = 0

        # Chop small imaginary parts
        imag_part = np.imag(complex_array)
        imag_part.setflags(write=True)
        imag_part[np.abs(imag_part) < threshold] = 0

        # Reconstruct the complex array with chopped parts
        kroneckerarray = real_part + 1j * imag_part
        
        

        return kroneckerarray


    def _CGC(self, idx_p, idx_q):
    
        """
        Calculates the CG coefficients for the product of the p  irrep times the q irrep. The ordering (1-based) is based in the character table obtained from GAP.
        
        Input:
        ----------
        
            - idx_p: int
                Index (1-based) of the p irrep according to character table. The ordering is based in the character table obtained from GAP.
            - idx_q: int
                Index (1-based) of the q irrep according to character table. The ordering is based in the character table obtained from GAP.
          
          
        Output:
        -------
        
            - CGCmatrix: numpy array.
                This numphy array contains the CG coefficients.
          
        """
        sp = self.dimensions_of_reps[idx_p-1]
        sq = self.dimensions_of_reps[idx_q-1]
        U = []
        
        
        #Calculate the CG coefficients from the matrix A()
        for idx_r in range(1, self.number_of_reps + 1):
            sr = self.dimensions_of_reps[idx_r-1]
            lmulti = self.getMultiplicity(idx_p, idx_q, idx_r)
            if lmulti == 0:
                continue

            lA = np.array([[self._A(idx_r, idx_p, idx_q, l+1, k+1) for k in range(sr)] for l in range(sr)])
            lk = 0
            ln = 1
            
            while ln <= lmulti:
                lc = np.array([np.array(self._getC(sr, sp, sq, ll+1, lA, lk)) for ll in range(sr)])
                U.append(lc)
                ln += 1

                li, lj, lk = self._pickTuple(sr, sp, sq, lA, lk)
                lk += 1
                
        U2 = np.array([item for sublist in np.array(U, dtype=object) for item in sublist])
        complex_array = np.transpose(U2)
        
        # Define a threshold for chopping small values
        threshold = 1e-5

        real_part = np.real(complex_array)
        real_part[np.abs(real_part) < threshold] = 0

        imag_part = np.imag(complex_array)
        imag_part.setflags(write=True)
        imag_part[np.abs(imag_part) < threshold] = 0

        CGCmatrix = real_part + 1j * imag_part
        
        return CGCmatrix


    def _CGCUnitary(self, idx_p, idx_q):
    
        """
        Calculates the CG coefficients for the product of the p  irrep times the q irrep. The ordering (1-based) is based in the character table obtained from GAP.

        Input:
        ----------
        
           - idx_p: int
                Index (1-based) of the p irrep according to character table. The ordering is based in the character table obtained from GAP.
           - idx_q: int
                Index (1-based) of the q irrep according to character table. The ordering is based in the character table obtained from GAP.
          
          
        Output:
        -------
        
            - CGCmatrixunitary: numpy array. A numpy array
                This numpy array contains the CG coefficients when the group is unitary.
          
        """
        
        
        sp = self.dimensions_of_reps[idx_p-1]
        sq = self.dimensions_of_reps[idx_q-1]
        U = [] #Define the matrix of CG coefficients
        # Define a threshold for chopping small values
        threshold = 1e-5

        #Calculate the CG coefficients from the matrix A()
        for idx_r in range(1, self.number_of_reps + 1):
            lUr = []
            sr = self.dimensions_of_reps[idx_r-1]
            lmulti = self.getMultiplicity(idx_p, idx_q, idx_r)
            if lmulti == 0:
                continue

            lAold = np.array([[self._A(idx_r, idx_p, idx_q, l+1, k+1) for k in range(sr)] for l in range(sr)])
            
            # Chop small numbers
            real_part = np.real(lAold)
            real_part[np.abs(real_part) < threshold] = 0
            imag_part = np.imag(lAold)
            imag_part.setflags(write=True)
            imag_part[np.abs(imag_part) < threshold] = 0
            lAold = real_part + 1j * imag_part
            
            
            # Compute the normalized coefficient lc
            lc = self._getCNormalized(sr, sp, sq, lAold)
            U.append(lc)
            
            for ln in range(2, int(lmulti.real) + 1):
                lAnew = np.array([
                [lAold[ll, lk] - np.dot(np.array([lc[ll]]).T, np.conj(np.array([lc[lk]]))) for lk in range(sr)]
                    for ll in range(sr)])#Aqui
                    
                # Chop small numbers
                real_part = np.real(lAnew)
                real_part[np.abs(real_part) < threshold] = 0
                imag_part = np.imag(lAnew)
                imag_part.setflags(write=True)
                imag_part[np.abs(imag_part) < threshold] = 0
            
                lAnew = real_part + 1j * imag_part
            
                lc = np.array(self._getCNormalized(sr, sp, sq, lAnew))
                
                U.append(lc)
                lAold = lAnew
        U2 = np.array([item for sublist in np.array(U, dtype=object) for item in sublist])
        complex_array = np.transpose(U2)

        

        # Chop small numbers
        real_part = np.real(complex_array)
        real_part[np.abs(real_part) < threshold] = 0
        imag_part = np.imag(complex_array)
        imag_part.setflags(write=True)
        imag_part[np.abs(imag_part) < threshold] = 0

        CGCmatrixunitary = real_part + 1j * imag_part

        return CGCmatrixunitary

    def getCGC(self, idx_p, idx_q):
        """
        Gets the CG coefficient from the CGC attribute. It also checks if a given CG coefficient has been computed before or not to avoid computing the same CG coefficients twice.
        
        Input:
        ----------
        
            - idx_p: int
                Index (1-based) of the p irrep according to character table. The ordering is based in the character table obtained from GAP.
            - idx_q: int
                Index (1-based) of the q irrep according to character table. The ordering is based in the character table obtained from GAP.
          
        Output:
        -------
        
            - cg_coeff_array_pq: numpy array
                A numpy array with the CG coefficients of the product p x q.
        
        """
        
        
        if not (1 <= idx_p <= len(self.dimensions_of_reps)):
            raise ValueError(
            f"Invalid irrep index: idx_p = {idx_p}. It must be between 1 and {len(self.dimensions_of_reps)} (inclusive)."
            )
        if not (1 <= idx_q <= len(self.dimensions_of_reps)):
            raise ValueError(
            f"Invalid irrep index: idx_q = {idx_q}. It must be between 1 and {len(self.dimensions_of_reps)} (inclusive)."
            )
            
        
        # Check if the Clebsch-Gordan coefficients have already been calculated
        if self.cg_coeff_calculated[idx_p-1, idx_q-1] != True:
            # Check if the group has unitary representations
            if self.unitary_representations:
                # Calculate the unitary Clebsch-Gordan coefficients
                self.cg_coeff[idx_p-1, idx_q-1] = self._CGCUnitary(idx_p, idx_q)
            else:
                # Calculate the general Clebsch-Gordan coefficients
                self.cg_coeff[idx_p-1, idx_q-1] = self._CGC(idx_p, idx_q)
            
            # Mark the coefficients as calculated
            self.cg_coeff_calculated[idx_p-1, idx_q-1] = True
        
        # Return the Clebsch-Gordan coefficients
        cg_coeff_array_pq = self.cg_coeff[idx_p-1, idx_q-1]
        return cg_coeff_array_pq

    def getCGCcc(self, idx_p, idx_q):
    
        """
        Gets the complex transpose of the matrix with the CG coefficients for the product of irreps p x q. If the group is not unitary, it gets the inverse matrix of the matrix with the CG coefficients. It also checks if this computation has been done before for a given product to avoid computing the same CG coefficients twice.
        
        Input:
        ----------
        
            - idx_p: int
                Index (1-based) of the p irrep according to character table. The ordering is based in the character table obtained from GAP.
            - idx_q: int
                Index (1-based) of the q irrep according to character table. The ordering is based in the character table obtained from GAP.
          
        Output:
        -------
        
            - cg_coeff_array_pq_cc: numpy array.
                A numpy array with the complex cojugate matrix (inverse matrix) of the CG coefficients of the product p x q for a unitary (non-unitary) group.
        
        """
        
        if not (1 <= idx_p <= len(self.dimensions_of_reps)):
            raise ValueError(
            f"Invalid irrep index: idx_p = {idx_p}. It must be between 1 and {len(self.dimensions_of_reps)} (inclusive)."
            )
        if not (1 <= idx_q <= len(self.dimensions_of_reps)):
            raise ValueError(
            f"Invalid irrep index: idx_q = {idx_q}. It must be between 1 and {len(self.dimensions_of_reps)} (inclusive)."
            )
        
        if self.unitary_representations:
            # Return the conjugate transpose of the Clebsch-Gordan coefficients
            cg_coeff_array_pq_cc = np.conjugate(np.transpose(self.getCGC(idx_p, idx_q)))
            return cg_coeff_array_pq_cc
        else:
            # Return the inverse of the Clebsch-Gordan coefficients
            cg_coeff_array_pq_cc  = np.linalg.inv(self.getCGC(idx_p, idx_q))
            return cg_coeff_array_pq_cc


    def getKroneckerProduct(self, idx_p, idx_q):
    
        """
        Gets the multiplicity of the all irreducible representations obtained in the decomposition of the product of irreps p x q.
        
        Input:
        ----------
        
            - idx_p: int
                Index (1-based) of the p irrep according to character table. The ordering is based in the character table obtained from GAP.
            - idx_q: int
                Index (1-based) of the q irrep according to character table. The ordering is based in the character table obtained from GAP.
          
        Output:
        -------
        
             - kronecker_product_array_pq:  numpy array
                A numpy array with the multiplicities of all the irreps in the product of the irreps p x q.
        
        """
        
        
        if not (1 <= idx_p <= len(self.dimensions_of_reps)):
            raise ValueError(
            f"Invalid irrep index: idx_p = {idx_p}. It must be between 1 and {len(self.dimensions_of_reps)} (inclusive)."
            )
        if not (1 <= idx_q <= len(self.dimensions_of_reps)):
            raise ValueError(
            f"Invalid irrep index: idx_q = {idx_q}. It must be between 1 and {len(self.dimensions_of_reps)} (inclusive)."
            )
        # Check if the Kronecker product has already been calculated
        if self.kronecker_product_calculated[idx_p-1, idx_q-1] != True:
            # Calculate the Kronecker product using the _generateKroneckerProduct function
            self.kronecker_product[idx_p-1, idx_q-1] = self._generateKroneckerProduct(idx_p, idx_q)
            # Mark the Kronecker product as calculated
            self.kronecker_product_calculated[idx_p-1, idx_q-1] = True
        
        # Return the Kronecker product
        kronecker_product_array_pq= self.kronecker_product[idx_p-1, idx_q-1]
        return kronecker_product_array_pq

    def getMultiplicity(self, idx_p, idx_q, idx_r):
    
        """
        Gets the multiplicity of the the r irrep obtained in the decomposition of the product of irreps p x q .
        
        Input:
        ----------
        
            - idx_p: int
                Index (1-based) of the p irrep according to character table. The ordering is based in the character table obtained from GAP.
            - idx_q: int
                Index (1-based) of the q irrep according to character table. The ordering is based in the character table obtained from GAP.
            - idx_r: int
                Index (1-based) of the r irrep according to character table. The ordering is based in the character table obtained from GAP.
          
        Output:
        -------
        
            - multuplicity_pq: int
                An integer which is the multiplicity of the irrep r in the product of the irreps p x q.
        
        """
        
        
        if not (1 <= idx_p <= len(self.dimensions_of_reps)):
            raise ValueError(
            f"Invalid irrep index: idx_p = {idx_p}. It must be between 1 and {len(self.dimensions_of_reps)} (inclusive)."
            )
        if not (1 <= idx_q <= len(self.dimensions_of_reps)):
            raise ValueError(
            f"Invalid irrep index: idx_q = {idx_q}. It must be between 1 and {len(self.dimensions_of_reps)} (inclusive)."
            )
        if not (1 <= idx_r <= len(self.dimensions_of_reps)):
            raise ValueError(
            f"Invalid irrep index: idx_r = {idx_r}. It must be between 1 and {len(self.dimensions_of_reps)} (inclusive)."
            )
        # Retrieve the Kronecker product for (p, q) and return the r-th element
        multuplicity_pq = round(self.getKroneckerProduct(idx_p, idx_q)[idx_r-1].real)
        return multuplicity_pq

    def _decomposeTensorProduct(self, idx_p, idx_q, vec):
        
        """
        Helper  function that takes a vector  in the product of irreps p x q, and partitions it into sub-vectors according to the multiplicities and dimensions of irreps appearing in the CG decomposition of p  x q.
        
        Input:
        ----------
        
            - idx_p: int
                Index (1-based) of the p irrep according to character table. The ordering is based in the character table obtained from GAP.
            - idx_q: int
                Index (1-based) of the q irrep according to character table. The ordering is based in the character table obtained from GAP.
            - vec: numpy array
                Matrix with the result of the tensor product p x q
        
          
        Output:
        -------
        
            - list_irreps: list
                A list with the different irreps of the product p x q obtained. Each entry corresponds to an irrep r in the decomposition p x q.
        
        """
    
        list_irreps = []
        lvec = vec
        
        for idx_r in range(1, self.number_of_reps + 1):
            lmult = self.getMultiplicity(idx_p, idx_q, idx_r)
            ltemp = []
            ldim = self.dimensions_of_reps[idx_r-1]
            
            for li in range(1, lmult + 1):
                lred = lvec[:ldim]
                if not self._VectorZeroQ(lred):
                    ltemp.append(lred)
                lvec = lvec[ldim:]
            
            list_irreps.append(ltemp)
        
        return list_irreps




    def _mJoin(self, list_X, list_Y):
        """
        Helper function for multiply_two. It joins different irreps in a single list of irreps.
        
        Input:
        ----------
        
            - list_X: list of lists of sympy symbols
                The first list to join.
            - list_y: list of lists of sympy symbols
                The second list to join.
           
        Output:
        -------
        
            - list_XY: list
                A list of lists of sympy symbols which contains the concatenation of list_X and list_Y.
        
        """
        
        if not list_X and not list_Y:  # If both are empty, return an empty list
            list_XY = []
            return list_XY
        if not list_X:  # If pX is empty, return pY
            list_XY = list_Y
            return list_XY
        if not list_Y:  # If pY is empty, return pX
            list_XY = list_X
            return list_XY
        list_XY = [px + py for px, py in zip(list_X, list_Y)]
        return list_XY # Otherwise, join the lists



    def getRepVector(self, idx_p, label_p):
    
        """
        Creates a list of lists with sympy symbols using label_p as a label. This list corresponds to a field transforming in the p irrep. The ordering (1-based) is based in the character table obtained from GAP.
        
        Input:
        ----------
        
            - idx_p: int
                Index (1-based) of the p irrep according to character table. The ordering is based in the character table obtained from GAP.
            - label_p: str
                string used to name the fields in this irrep.
        
        Output:
        -------
        
            - repVector: list of sympy symbols
                A list of sympy symbols with a field in a irrep p.
        
        """
        
        
        if not (1 <= idx_p <= len(self.dimensions_of_reps)):
            raise ValueError(
                f"Invalid irrep index: idx_p = {idx_p}. It must be between 1 and {len(self.dimensions_of_reps)} (inclusive)."
            )
                        
        repVector = [[] for _ in range(self.number_of_reps)]  # Retrieve the representation vector
        pRep = int(idx_p)
        if pRep > 0:
            repVector[pRep - 1] = [self._genLabels(self.dimensions_of_reps[pRep-1], label_p)]
        
        return repVector


    def _genLabels(self, dim_p, label_p):
    
        """
        Creates the different labels for the getRepVector method.
        
        Input:
        ----------
        
        
           - dim_p: int
               the dimension of the irrep
           - label_p: str
               string used to name the fields in this irrep. For instance, if label_p = "x", then they will be named as sp.Symbolx("xr1"), sp.Symbolx("xr2"), ...
        
        Output:
        -------
        
            - labels: list of sympy symbols
                A list of sympy symbols to generate the field that transforms in a given irrep.
        
        """
        labels = []
        for lf in range(1, dim_p + 1):
            var_name = f"{label_p}r{lf}"
            symbol = sp.Symbol(var_name)
            labels.append(symbol)
        return labels

    def _multOneRep(self, idx_li, idx_lj, list_A, list_B):
        
        """
        Calculates the product of the index idx_li irrep of the list list_A with the idx_lj-th irrep of the list list_B. Here list_A and list_B can be fields that only transform in one irrep. Or they can be lists with different irreps.
        
        Input:
        ----------
        
            -  idx_li: int
               Index (1-based) of the li irrep of the list_A. The ordering is based in the character table obtained from GAP.
            - idx_lj: int
               Index (1-based) of the li irrep of the list_A. The ordering is based in the character table obtained from GAP.
            - list_A : list of sympy symbols
               A list that contains different irreps.
            - list_B : list of sympy symbols
               A list that contains different irreps.
        
        Output:
        -------
        
            - mult: list of sympy symbols
                A list of sympy symbols. This list contains the product of the li irrep of list_A with the lj irrep of list_B.
        
        """

        def process_y(y): #Helper functions to compute the tensor product between two irreps
            def process_x(x):
                tensor_product = np.kron(x, y) # Aqui
                transformed_tensor = np.dot(self.getCGCcc(idx_li, idx_lj), tensor_product)
                return self._decomposeTensorProduct(idx_li, idx_lj, transformed_tensor)
            
            return list(map(process_x, list_A[idx_li-1]))  # Adjusted for 0-based indexing
        
        mult = list(map(process_y, list_B[idx_lj-1]))
        return mult # Adjusted for 0-based indexing


    def _eliminate_small_real_imag(self, expr):
        """
        Helper function to eliminate small real and imaginary parts in sympy expressions
        
        Input:
        ----------
    
        - expr: sympy Symbol
        

        Output:
        -------
            - expr_chopped: sympy symbol.
                A sympy Symbol with chopped real and imaginary numbers.
        
        """
        def replacement(f): #Helper function to chop a float
            real_part = f.as_real_imag()[0]
            imag_part = f.as_real_imag()[1]

            new_real = 0 if abs(real_part) < 1e-5 else real_part
            new_imag = 0 if abs(imag_part) < 1e-5 else imag_part
            
            return new_real + sp.I*new_imag
        
        expr_chopped = expr.replace(lambda e: e.is_Float, replacement)
        return expr_chopped



    
    def _flatten_once(self, lst):
        """
        Helper function that acts as Mathematica's Flatten level 1.
        
        Input:
        ---------
        
            - lst: A list
                A list which each element is a list
           
           
        Output:
        -------
        
            - flatted_lst: list.
                A list where all the lists inside lst were concatanated.
        """
        flatted_lst = list(itertools.chain.from_iterable(lst))
        return flatted_lst

    def _multiply_two(self, list_A, list_B):
    
        """
        It gives a list with the different irreps obtained from the tensor product between the vector list_A with list_B.
        
        
        Input:
        ----------
        
            - list_A: list of lists with sympy symbols
                This is the first list with irreps to use in the tensor multiplication.
            - list_B: list of lists with sympy symbols
                This is the second list with irreps to use in the tensor multiplication.
           
           
        Output:
        -------
        
            - product:A list of lists with sympy symbols.
                This list contains all the irreps resulting from the product of list_A x list_B.
        
        """
        lres = [[] for _ in range(self.number_of_reps)]
        
        if not isinstance(list_A, list):
            raise ValueError("Expected a list for 'list_A', but got a non-list object.")

        if len(lres) != len(list_A):
            raise ValueError(f"Length mismatch: expected {len(lres)} elements, got {len(list_A)}.")

        if not all(isinstance(sublist, list) for sublist in list_A):
            raise ValueError("Each element of 'list_A' must be a list.")
            
        if not isinstance(list_B, list):
            raise ValueError("Expected a list for 'list_B', but got a non-list object.")

        if len(lres) != len(list_B):
            raise ValueError(f"Length mismatch: expected {len(lres)} elements, got {len(list_B)}.")

        if not all(isinstance(sublist, list) for sublist in list_B):
            raise ValueError("Each element of 'list_B' must be a list.")
            
        
        
        if list_A == lres:
            product = list_B
            return product
        
        if list_B == lres:
            product = list_A
            return product
        
        for li in range(len(list_A)):
            for lj in range(len(list_B)):
                if li > lj:
                    ltemp = self._multOneRep(lj + 1, li + 1, list_B, list_A)
                else:
                    ltemp = self._multOneRep(li + 1, lj + 1, list_A, list_B)
                
                flattened_temp = self._flatten_once(ltemp)# Flatten by one level
                if not not flattened_temp:
                    for k in range(len(flattened_temp)):
                           lres = self._mJoin(flattened_temp[k],lres)
            

        result = [[[self._eliminate_small_real_imag(expr) for expr in subsublist] for subsublist in sublist] for sublist in lres]
        
        # Create a set to track unique expressions
        product  = []
        for i in range(len(result)):
            unique_exprs = []
            for expr in result[i]:
                if not any(expr[0] - unique[0] == 0 for unique in unique_exprs):
                    unique_exprs.append(expr)
            product.append(unique_exprs)
        

        return product
        
    def multiply(self,list_fields):
        """
        It gives a list with the different irreps obtained from the tensor product of the fields included in the input.
        
        Inputs:
        -------
        
            - list_fields: list of lists with sympy symbols
                This list include all the fields to be multiplied
                  
        Outputs:
        --------
        
        
             - res: list of lists with sympy symbols
                Irrep decomposition of the product of the fields in list_fields
             
        """
    
        res = list_fields[0]
        for i in range(1, len(list_fields)):
            res = self._multiply_two(res, list_fields[i])
        return res
        


    def extractOneRep(self, vec_rep, idx_p):
        """
        It extracts one the index (1-base) p irrep from the list of irreps vec_rep. The ordering of the irreps (1-based) is based in the character table obtained from GAP.
        
        Input:
        ----------
        
             - vec_rep: list of lists of symbols
                This is the list of irreps from which you will extract only one of the irreps.
             - idx_p:  int
                This is the index (1-based) of the p irrep that you want to extract.
            
        Output:
        -------
        
              - oneRep_list: A list of lists of symbols with the p irrep of the vector vec_rep.
                This list will contain mostly empty lists except from the irrep you extracted.
           
        """
        
        lres = [[] for _ in range(self.number_of_reps)]
        
        
        if not isinstance(vec_rep, list):
            raise ValueError("Expected a list for 'vec_rep', but got a non-list object.")

        if len(lres) != len(vec_rep):
            raise ValueError(f"Length mismatch: expected {len(lres)} elements, got {len(vec_rep)}.")

        if not all(isinstance(sublist, list) for sublist in vec_rep):
            raise ValueError("Each element of 'vec_rep' must be a list.")
            
            
        if not (1 <= idx_p <= len(self.dimensions_of_reps)):
            raise ValueError(
            f"Invalid irrep index: idx_p = {idx_p}. It must be between 1 and {len(self.dimensions_of_reps)} (inclusive)."
            )
        
        idx_ppython = idx_p - 1
        oneRep_list = [elem if idx == idx_ppython else [] for idx, elem in enumerate(vec_rep)]
        return oneRep_list


    def _flatten(self, nested_list):
        """
        It flattens like mathematica.
        
        Input:
        ----------
        
             - nested_list: list of lists
                This list will be flattened like Mathematica
            
        Output:
        -------
        
              - flat_list: A list.
                A flattened list
        """
        flat_list = []
        for item in nested_list:
            if isinstance(item, list):
                flat_list.extend(self._flatten(item))
            else:
                flat_list.append(item)
        return flat_list



"""
PyDiscrete Subpackage
=====================

Provides:
- `Group`: main class interfacing with GAP to compute group information for particle model-building
"""

from .PyDiscrete import Group

__all__ = ["Group"]
__version__ = "0.2.0"



import numpy as np
import sympy as sp
import json
import os
from itertools import product
from itertools import permutations
from FlavorBuilder.PyDiscrete import Group
import random
from sympy import symbols, expand, Symbol
from collections import defaultdict
from collections import Counter
from numbers import Real
import re







class Model2Mass():

    def __init__(self, group, model_file = None,  model = None, vev_dict = None, non_flavon_particles = 11, particle_list_grouping = None):
    
    
        """
        
        Inputs:
        --------
            - group: PyDiscrete group object.
                A group obtained by the PyDiscrete package.
                
            - model_file: str
                A string containing the path to the file containing the model.
                
            - model: dictionary
                A dictionary containing all the necessary details to extract the lagrangian and mass matrices. See the tutorial for details.
                
            - vev_dict: dictionary (Optional. The default is set by self._setFlavonVeVsdict(). See the docstring therein)
                A dictionary that will specify the vevs for the different vevs.
                
            - non_flavon_particles: int (Optional. Set to 11 by default)
                An integer specifying the number of non-flavon particles.
                
            - particle_list_grouping: dictionary (Optional. The default is set by _setParticleListGrouping(). See the docstring therein)
                A list of lists  with two strings, specifying the left leptons, right leptons, ... and so on.
                
                
                
                
        Atrributes:
        ----------
            - model_file: str
                Path to the model_file
            - model: dictionary
                Model dictionary with all the necessary specifications to build a lagrangian. See the tutorial for more details.
            - abelian_group_dimensions: array
                List with the dimensions of the ZN symmetries. It is obtained from the "ZN_symmetries" key of the model dictionary.
            - parametersSet: dictionary
                Dictionary with the values that were set in the model. For instance, the flavon vevs, or the vevs of the higgs to ignore them when calculating the dimensionless mass matrix.
            - flavonvevs: list of sympy symbols
                Flavon sympy symbols
            - list_of_fields_sympy_symbols: list of sympy symbols
                Sympy symbols for all the fields considered
            - list_of_ALL_parameters_in_lag: list of sympy symbols
                Sympy symbols for all the paramaeters in the lagrangian (denoted by alpha{n})
            - numberofAbelianSymmetries: int
                Number of Abelian Symmetries
            - NumberOfTotalFlavons: int
                Number of flavons with a non-zero array in the "representations" key of the model dictionary
            - particle_dict: dictionary
                Dictinary with all the particles in a notation digestible for PyDiscrete
            - non_flavon_particles: int
                Number of particles that are non-flavon

        Methods:
        -----------
        
            - fields_for_discrete(data_model):
                Produce matrix of field representations in format ingestible by PyDiscrete.
                

                
            - get_lagrangian(max_terms = 50, allOperators = None):
                Get the lagrangian and a list of parameters with its respective term in the lagrangian. The max_terms sets the number of terms in the lagrangian after which the computations stops. allOperators is a list of lists with strings that indicates which terms should be compueted for the lagrangian. See the tutorial for more details.
            
            - ParametersInAMatrix(mat):
                Gets the parameters in a sympy matrix mat
            
            - get_rank_mass_matrix(mass_matrix, parameters):
                Gets the rank of mass_matrix. A list of parameters obtained from get_lagrangian must be provided
            
            - count_eff_flavons(lagrangian):
                Gives the number of flavons that appear in the lagrangian and a list with the name of the flavons that do not appear in the lagrangian
            
            - get_mass_matrix(lagrangian ,parameters, left_fields, right_fields = None, higgs_fields = None,  factor_out = True):
                Gets the mass matrix of the lagrangian. It needs the parameters output from get_lagrangian. Also, it needs a list of left_fields, right_fields and higgs_fields to perform the derivatives to calculate the mass matrix. The factor_out input, if true, factors out one of the parameters and sets it to one in the model. This is useful since it makes the fitting process faster.
        
        
        """
       
       
        if group == None:
           raise ValueError("You must provide a group from PyDiscrete")
        self.group = group
        
        if (model_file is None) and (model is None):
            raise ValueError("You must provide a model file or a model.")
     
        self.model_file = model_file
        self.model = model
        
        
        self.abelian_group_dimensions = 0
        self.parametersSet = {}
        self.flavonvevs = []
        self.list_of_fields_sympy_symbols = []
        self.list_of_ALL_parameters_in_lag = []
        self.numberofAbelianSymmetries = 0
        self.NumberOfTotalFlavons = 0
        self.particle_dict = {}
        
        self.non_flavon_particles = non_flavon_particles
        
        
        
        self.model = self._read_model()
        self._setParticleListGrouping(particle_list_grouping)
        self._setFlavonVeVsdict(vev_dict)
        


        

    def _read_model(self):
        '''
        Format data for PyDiscrete. It saves the information of the model_file into python variables to be handled easily

        Output:
        -------
            - data_model: dictionary
                This dictionary contains the particle representation information obtained from the model_file with absent particles removed.
        '''
        if self.model_file is not None:
            with open(self.model_file, 'r') as file:
                data_model0 = json.load(file)
        if self.model is not None:
            data_model0 = self.model
        if (self.model is not None) and (self.model_file is not None):
            with open(self.model_file, 'r') as file:
                tempdata_model0 = json.load(file)
            if self.model != tempdata_model0:
                raise  ValueError("You gave both a model and a model file but they are not equal. They should be equal.")
            
        # Extract particles, representations, and associations
        particles = data_model0["particles"]
        representations = data_model0["representations"]
        fixed_flavon_vevs = data_model0["fixed_flavon_vevs"]
        ZN_symmetries = data_model0["ZN_symmetries"]
        self.numberofAbelianSymmetries = len(representations[0]) - self.group.number_of_reps
        self.abelian_group_dimensions = ZN_symmetries
        self.parametersSet = {}
             
        # Remove all zero entries
        data_model = {
            'representations': [],
            'associations': [],
            'particles': [],
            'fixed_flavon_vevs': [],
            'ZN_symmetries': []
        }
        rep_array = np.array(representations)
        all_zero_rows = np.all(rep_array == 0, axis=1)
        mask = np.ones(rep_array.shape[0], dtype=bool)
        mask[all_zero_rows] = False
        data_model['representations'] = rep_array[mask].tolist()
        data_model['associations'] = data_model0['associations']
        data_model['particles'] = np.array(particles)[mask].tolist()
        data_model['ZN_symmetries'] = ZN_symmetries

                
        rep_array_vevs = np.array(fixed_flavon_vevs)
        self.NumberOfTotalFlavons = len(rep_array_vevs)
        all_zero_rows_flavonsVevs = all_zero_rows[self.non_flavon_particles:]
        mask_flavonsVevs = np.ones(np.array(rep_array_vevs).shape[0], dtype=bool)
        mask_flavonsVevs[all_zero_rows_flavonsVevs] = False
        
        
        data_model['fixed_flavon_vevs'] = rep_array_vevs[mask_flavonsVevs].tolist()
        

        return data_model

    def fields_for_discrete(self, data_model):
        '''
        Produce matrix of field representations in format ingestible
        by PyDiscrete.

        Inputs:
        -------
        
            - data_model: Dictionary
                 This is a dictionary of particle representation information.

        Outputs:
        --------
        
            - fieldsForDiscrete: A matrix of field representations in
                                 PyDiscrete friendly format.
        '''
        
        
        SeesawForDiscrete_lst = self._SeesawForDiscrete(data_model) #This prepares the MSSM + 3 right-handed neutrinos for PyDiscrete
        flavonsForDiscrete_lst = self._read_flavons(data_model) #This prepares the flavons for PyDiscrete
        
        
        fieldsForDiscrete = SeesawForDiscrete_lst + flavonsForDiscrete_lst #This list will include all MSSM fields + 3 right-handed neutrinos + flavons

        return fieldsForDiscrete
        
    def _SeesawNpletsWithRow(self, data_model):
        """
        Helper function to group the MSSM and the 3 right-handed neutrinos appropiately. For instance, if the three left SU(2) doublets belong to a triplet, they
        need to be grouped together.
        
        Inputs:
        -------
        
            - data_model: Dictionary
                 This is a dictionary of particle representation information.
                 
        Outputs:
            - SeesawNpletsWithRow: list
                 This is a list that encodes the information of the reps for the MSSM + 3 right-handed neutrinos. It also groups together particles that belong to the same multiplet.
        
        """
        
        
        # List to store the results
        SeesawNpletsWithRow = []
        already_grouped = set()


        # Loop through the first eleven particles which correspond to the MSSM + 3 right-handed neutrinos
        for i, particle in enumerate(data_model['particles'][:self.non_flavon_particles]):
            
            # Skip if the particle is already grouped
            if particle in already_grouped:
                continue
            
            representation = data_model['representations'][i]
            

            rep_num = np.where(representation)[0][0] + 1
            
            # Check if the irrep is a singlet
            if rep_num <= self.group.NumberOfSinglets:
            
                # Add single particle entry (particle, A4 rep, ZN rep)
                SeesawNpletsWithRow.append([[particle], rep_num, representation[self.group.number_of_reps:self.group.number_of_reps+self.numberofAbelianSymmetries]])
                already_grouped.add(particle)
            else:
                # Group particles with the same non-zero entry
                group = [particle]
                already_grouped.add(particle)
                in_group = np.where(np.array(data_model['associations']) == data_model['associations'][i])[0][1:]
                for k in in_group:
                    other_particle = data_model['particles'][k]
                    other_representation = data_model['representations'][k]
                    group.append(other_particle)
                    already_grouped.add(other_particle)

                # Add the field to the list SeesawNpletsWithRow depending on whether they are singlets or not
                if len(group) > 1: #Non-singlets
                    SeesawNpletsWithRow.append([group, rep_num, representation[self.group.number_of_reps:self.group.number_of_reps+self.numberofAbelianSymmetries]])
                else: #Singlet
                    SeesawNpletsWithRow.append([[particle], rep_num, representation[self.group.number_of_reps:self.group.number_of_reps+self.numberofAbelianSymmetries]])
                    
        return SeesawNpletsWithRow
        
        
    def _SeesawForDiscrete(self, data):
        """
        Helper function to make  the MSSM and the 3 right-handed neutrinos digestible for PyDiscrete.
        
        Inputs:
        -------
        
            - data_model: Dictionary
                 This is a dictionary of particle representation information.
                 
        Outputs:
            - MSSM3NForDiscrete: list
                 This is a list that encodes the information of the reps for the MSSM + 3 right-handed neutrinos digestible for PyDiscrete format.
        
        """
         
        SeesawNpletsWithRow = self._SeesawNpletsWithRow(data)
        MSSM3NForDiscrete = []
        for nplet in SeesawNpletsWithRow: #Add MSSM fields + 3 right-handed neutrinos to fieldsForDiscrete
            repNumber = int(nplet[1])
            if repNumber <= self.group.NumberOfSinglets:
                for j in range(len(nplet[0])):
                    fieldname = nplet[0][j]
                    var_name = f"{fieldname}"
                    symbol = sp.Symbol(var_name)
                    self.list_of_fields_sympy_symbols.append(symbol)
                    lres = [[] for i in range(self.group.number_of_reps)]  # Retrieve the representation vector
                    lres[repNumber - 1] = [[symbol]]
                    MSSM3NForDiscrete.append([lres,nplet[-1:][0]])
            else: #Create n-plets for flavons and add them to fieldsForDiscrete
                lres = [[] for i in range(self.group.number_of_reps)]  # Retrieve the representation vector
                lres2 =[]
                for j in range(len(nplet[0])):
                    fieldname = nplet[0][j]
                    var_name = f"{fieldname}"
                    symbol = sp.Symbol(var_name)
                    self.list_of_fields_sympy_symbols.append(symbol)
                    lres2.append(symbol)
                lres[repNumber - 1] = [lres2]
                MSSM3NForDiscrete.append([lres,nplet[-1:][0]])
                
        return MSSM3NForDiscrete
        
        
    def _setFlavonVeVsdict(self, vev_dict):
        """
        This function sets the identification of the different flavon vevs you consider
        
        Inputs:
        ------
        
            - vev_dict: Dictionary
                This is a dictionary between integer numbers and flavon vevs. If none, it goes to the defaul list given by:
                {1 :  np.array([1]), 2: np.array([1, 0, 0]), 3: np.array([1, 1, 1]), 4: np.array([0, 1, 0]), 5: np.array([0, 0, 1]), 6: np.array([0, 1, -1]) , 7 : np.array([1, -(-1)**(1/3), (-1)**(2/3)]) , 8: np.array([1, (-1)**(2/3), -(-1)**(1/3)]) }
                
        Outputs:
        --------
        
        None
        
        """
        if vev_dict == None:
            self.vev_dict =  {1 :  np.array([1]), 2: np.array([1, 0, 0]), 3: np.array([1, 1, 1]), 4: np.array([0, 1, 0]), 5: np.array([0, 0, 1]), 6: np.array([0, 1, -1]) , 7 : np.array([1, -(-1)**(1/3), (-1)**(2/3)]) , 8: np.array([1, (-1)**(2/3), -(-1)**(1/3)]) }
        else:
            self.vec_dict = vev_dict
            
        if not all(isinstance(key, int) for key in self.vev_dict.keys()):
            raise ValueError("All the keys must be integers.")
        if not all( (isinstance(val, np.ndarray) and (np.issubdtype(val.dtype, np.floating) or np.issubdtype(val.dtype, np.integer) or np.issubdtype(val.dtype, np.complexfloating))) or (isinstance(val, list) and all(isinstance(x, (float, int, complex)) for x in val)) for val in self.vev_dict.values() ):
            raise ValueError("All values must be lists or NumPy arrays containing only floats")
            
            
            
    def _setParticleListGrouping(self, particle_list_grouping):
        
        """
        
        It sets the attribute particle_list_grouping in the model2mass class.
        Inputs:
        -------
            - particle_list_grouping: list of strings
                 This is a list of particles that will be used to group them to contruct the lagrangian terms. If particle_list_grouping = None, it gives the default [["leftLeptons", "L"], ["rightLeptons", "E"], ["rightNeutrinos", "N"],  ["flavons", "phi"], ["higgsup", "Hu"], ["higgsdown", "Hd"]]

        """
        
        
        if particle_list_grouping == None:
            self.particle_list_grouping = [["leftLeptons", "L"], ["rightLeptons", "E"], ["rightNeutrinos", "N"],  ["flavons", "phi"], ["higgsup", "Hu"], ["higgsdown", "Hd"]]
        
        else:
            if not all( isinstance(elem, list) and len(elem) == 2 and all(isinstance(sub_elem, str) for sub_elem in elem) for elem in particle_list_grouping ):
                raise ValueError("All elements of particle_list_grouping must be lists of exactly two strings.")
                
            self.particle_list_grouping = particle_list_grouping
        
        
             

    def _assignFixedVevs(self, fixed_flavon_vevs_index, repPosition, symbolr ):
        """
            This is a helper function that saves the fixed vevs selected by the user into the attribute ""parametersSet"". This attributed is later used when evaluating the mass matrices
            
        Inputs:
        -------
            - fixed_flavon_vevs_index: int
                This is the index that determines from the attribute "vev_dict" which vev will be assigned to the flavon
            - repPosition: int
                This is the index of the irrep for the flavon to be assigned a vev. It should follow the conventions by PyDiscrete.
            - symbolr : sympy symbol
                This is the sympy symbol that will be used to assign the name of this vev.
        """
        flavonInPydiscrete = self.group.getRepVector(repPosition, symbolr)
        dimOfRep = self.group.dimensions_of_reps[repPosition - 1] # the -1 is to follow the 0-index notation in python
        dimOfVeV = len(self.vev_dict[fixed_flavon_vevs_index])
        
        if  not dimOfVeV == dimOfRep:
            raise ValueError("The dimension that you selected for the flavon does not match the assigned vev in the dictionary you choose")
        
        for i in range(dimOfRep):
            self.parametersSet[self.group._flatten(self.group.getRepVector(repPosition, symbolr))[i]] = self.vev_dict[fixed_flavon_vevs_index][i]
        
        return None
        
        
    def _read_flavons(self, data_model):
        """
        Helper function to make flavons be part of a list digestible for PyDiscrete. It also sets the flavon vevs and saves their value in the class attribute parametersSet.
        
        Inputs:
        -------
        
            - data_model: Dictionary
                 This is a dictionary of particle representation information.
                 
        Outputs:
            - flavonsForDiscrete: list
                 This is a list that encodes the information of the reps for flavons digestible for PyDiscrete format.
        
        """
        flavonsForDiscrete = []
        
        flavonsinput = []
        for i in range(self.non_flavon_particles, len(data_model['particles'])):
            flavonsinput.append([data_model['particles'][i], data_model['representations'][i]])

        flavonvevs = []


        # Define the function to process each row
        for i in range(len(flavonsinput)):
            row = flavonsinput[i]
            symbol = row[0]
            symbolr = symbol + 'r'
            values = row[1]
            repPosition = np.where(values)[0][0] + 1
            self.list_of_fields_sympy_symbols.append(sp.Symbol(symbolr))

            # Construct the code string
            flavonsForDiscrete.append([self.group.getRepVector(repPosition, symbolr), row[1][self.group.number_of_reps:self.group.number_of_reps+self.numberofAbelianSymmetries]])
            flavonvevs.append(self.group.getRepVector(repPosition, symbolr))
            
            
            #Call _assignFixedVevs to assign vecs from vev_dict for later use
            fixed_flavon_vevs_index= data_model['fixed_flavon_vevs'][i]
            self._assignFixedVevs(fixed_flavon_vevs_index, repPosition, symbolr )
            
        
        self.flavonvevs = self.group._flatten(flavonvevs)
        
            
        return flavonsForDiscrete

    # Function to scan symbols inside each row
    def _scanRow(self, row, particle_dict, particle_list_grouping):
        '''
        Identify particles present in each row and assign to correct list in particle dictionary. It will assign the particle to the correct entry in the particle dictionary. This entry can be "left leptons" for the left SU(2) doublets leptons, "rightLeptons" for right SU(2) singlet leptons, "rightNeutrinos" for right-handed neutrinos, "higgsup" for Up higgs  , "higgsdown" for the down higgs and "flavons" for flavons.

        Inputs:
            - row: list.
                A list of representations for a single particle.
            - particle_dict: dictionary
                Dictionary categorizing particles.

        Outputs:
            - particle_dict: dictionaty
                Updated with information from row.
        '''
        
        
        
        # Flatten the row
        flattenedRow = [item for sublist in row for item in sublist]
        # Convert to strings (if not already)
        symbols = [str(item) for item in flattenedRow]

        
        # Append to the appropriate list based on the symbol content
        for i in range(len(particle_list_grouping)):
            if any(particle_list_grouping[i][1] in symbol for symbol in symbols):
                particle_dict[particle_list_grouping[i][0]].append(row)
        
        return particle_dict

    def _get_particle_dict(self, fieldsForDiscrete):
        '''
        Construct particle dictionary from fields. By default, it groups all left SU(2) doublets leptons, right SU(2) singlet leptons,
        right-handed neutrinos, higgs up , higgs down and flavons in a particle dictionary. It can also be made customary

        Input:
            - fieldsForDiscrete: list of sympy symbols
                 A list of sympy field representations in PyDiscrete friendly format.
        Output:
            - particle_dict: dictionary
                 Dictionary categorizing particles with their representations.
        '''

        
        particle_dict = {key[0]: [] for key in self.particle_list_grouping}
        
        # Scan each row in the fields
        for row in fieldsForDiscrete:
            particle_dict = self._scanRow(row, particle_dict, self.particle_list_grouping)
        return particle_dict
  
    
    # Function to return unique indices and indices of repeated lists
    def _sort_symbols(self, original_list):
        """
        Helper function that sorts symbols from a list
        
        Inputs:
        -------
            - original_list: list of sympy symbols
        
        Outputs:
        --------
            - sorted_list: sorted list of sympy symbols
        
        """
        unique_elements = []
        repeated_elements_dict = {}
    
        for idx, element in enumerate(original_list):
            if original_list.count(element) == 1:
                unique_elements.append(idx)
            else:
                if str(element) not in repeated_elements_dict:
                    repeated_elements_dict[str(element)] = [idx]
                else:
                    repeated_elements_dict[str(element)].append(idx)
    
        result = [unique_elements]
        for indices in repeated_elements_dict.values():
            result.append(indices)
        
        sorted_list = []
        for i in range(len(result)):
            temp_list = [original_list[j] for j in result[i]]
            sorted_list = sorted_list + temp_list
    
        return sorted_list

                                
    def _unique_combs_fields(self, list_fields, repeated_fields):
        '''
        This is a helper function that helps the multiplication in the lagrangian to not make repetitions.
        
        Inputs:
        -------
            - list_fields: list of sympy symbols:
                This list is the list of fields for each category.
            
            
            - repeated_fields: list of integers
                This list a list of integers where all the elements with the same integer will be grouped together since this means they are the same field.
            
        Outputs:
        --------
            - valid_tuples: list of tuples
                This list includes the valid indices such that the products in the lagrangian do not repeat.
        
        '''
        
        # Build index ranges for each list
        index_ranges = [range(len(f)) for f in list_fields]

        # Identify groups from repeated_fields
        group_to_indices = defaultdict(list)
        for idx, tag in enumerate(repeated_fields):
            if tag != 0:
                group_to_indices[tag].append(idx)

        # Generate full list of tuples
        all_index_tuples = product(*index_ranges)



        #Filter combinations that satisfy group-wise monotonicity
        valid_tuples = []
        for idx_tuple in all_index_tuples:
            is_valid = True
            for group_indices in group_to_indices.values():
                group_values = [idx_tuple[i] for i in group_indices]
                if group_values != sorted(group_values):
                    is_valid = False
                    break
            if is_valid:
                valid_tuples.append(idx_tuple)

        
        return valid_tuples
        
    def _operatorToLag(self, list_fields, lagrangian, parameters, n_terms, coeff, max_terms, repeated_fields):
    
        '''
        Construct lagrangian and parameters for the NN operator

        Input:
        -----
            - list_fields: list of lists of sympy symbols
                 Each element is a type of particle (flavons, right-handed neutrinos) category and within each category there are the irreps of those fields.
            - lagrangian: expression of sympy symbols
                 The lagrangian to which we will add the result of the operator "NN".
            - parameters: list tuples.
                 Each tuple is of the form (alpha_k, coeff, term_k) where alpha_k is the coefficient of the term in the lagrangian. Thus, the lagrangian looks like lag = \sum_k \alpha_k * coeff * term_k. Coeff encodes how much supressed is this term due to higher-order powers of \phi/\Lambda.
            - n_terms: int
                 Number of terms computed so far for the lagrangian.
            - max_terms: int
                 Maximum number of terms allowed in the lagrangian.
            - k: int
                 The index of the term at which the computation of the invariant term  starts.
            - repeated_fields: list of indices
                 This is a list that indicates which of the elements in list_fields are repeated

        Output:
        -------
            - lagrangian: expression of sympy symbols.
                  The input lagrangian  + the terms from the operator obtained from multiplying the fields in list_fields
            - parameters: list tuples.
                  The parameters input list with the new terms appended.
            - n_terms: int
                  The input n_terms + the number of terms computed in this function

                  
                  
                  
        '''
    
        valid_tuples = self._unique_combs_fields(list_fields, repeated_fields)
        
        for idx_tuple in valid_tuples:
            if n_terms > max_terms:
                break
            mod_condition = [int(sum((list_fields[nn][idx_tuple[nn]][1][jjj] - 1) for nn in range(len(list_fields)))) % self.abelian_group_dimensions[jjj] == 0 for jjj in range(self.numberofAbelianSymmetries)]

            if all(mod_condition):
                # Get invariant terms by calling the group multiplication from PyDiscrete

                sorted_terms = self._sort_symbols([list_fields[nn][idx_tuple[nn]] for nn in range (len(list_fields))])
                invariant_terms = self.group._flatten(self.group.extractOneRep(self.group.multiply([sorted_terms[nn][0] for nn in range(len(sorted_terms))]), 1))
                                    
                
                if invariant_terms:  # Check if invariant_terms is not empty
                    for nnn in range(len(invariant_terms)):
                        alpha_k = sp.symbols(f'alpha{n_terms + 1}')
                        var_name = f'alpha{n_terms + 1}'
                        symbol = sp.Symbol(var_name)
                        lagrangian += coeff * alpha_k * invariant_terms[nnn]
                        self.list_of_ALL_parameters_in_lag.append(alpha_k)
                        parameters.append([alpha_k, coeff, invariant_terms[nnn], [sorted_terms[nn] for nn in range(len(sorted_terms))]])
                        n_terms += 1
        return lagrangian, parameters, n_terms



    def _label_repeated_fields(self, list_fields):
        """
        This is a helper function that gives the correct list of integers for repeeated fields. It is used in get_lagrangian.
        
        Inputs:
        -------
                - list_fields: list of strings
                     List with the fields to obtain the repeated list
        
        Outputs:
        --------
                - list_repeated_fields: list of integers
                      List that indicates which fields are repeated.
        
        
        """
        counts = Counter(list_fields)       # Count each string's occurrences
        list_repeated_fields = [None] * len(list_fields)   # Output list
    
        repeat_label = 1                 # Start numbering repeated types from 1
        repeat_map = {}                  # Map repeated string → assigned label
    
        for i, s in enumerate(list_fields):
            if counts[s] == 1:
                list_repeated_fields[i] = 0  # Unique string
            else:
                if s not in repeat_map:
                    repeat_map[s] = repeat_label
                    repeat_label += 1
                list_repeated_fields[i] = repeat_map[s]
    
        return list_repeated_fields
    
    def _setallOperators(self, allOperators = None):
        """
        This is a helper function which saves the list of all operators that the user wants to compute
        
        Inputs:
        -------
              - allOperators: list of list with strings and coefficients
                   This is the list that contains all the terms you want to be considered in the Lagrngian. The first element of each entry should be a list with strings and the second a float which sets the supression coefficient. If None, it sets to the default:
                               [
                        [["rightNeutrinos", "rightNeutrinos"], 1],
                        [["flavons", "rightNeutrinos", "rightNeutrinos"], 0.1],
                        [["flavons", "flavons", "rightNeutrinos", "rightNeutrinos"], 0.01],
                        [["rightNeutrinos", "leftLeptons", "higgsup"], 1],
                        [["rightNeutrinos", "leftLeptons", "higgsup", "flavons"], 0.1],
                        [["leftLeptons", "rightLeptons", "higgsdown"], 1],
                        [["leftLeptons", "rightLeptons", "higgsdown", "flavons"], 0.1]
                    ]
                   
        Outputs:
        --------
              - None
        
        """
        
        if allOperators == None:
            self.allOperators = [
                        [["rightNeutrinos", "rightNeutrinos"], 1],
                        [["flavons", "rightNeutrinos", "rightNeutrinos"], 0.1],
                        [["flavons", "flavons", "rightNeutrinos", "rightNeutrinos"], 0.01],
                        [["rightNeutrinos", "leftLeptons", "higgsup"], 1],
                        [["rightNeutrinos", "leftLeptons", "higgsup", "flavons"], 0.1],
                        [["leftLeptons", "rightLeptons", "higgsdown"], 1],
                        [["leftLeptons", "rightLeptons", "higgsdown", "flavons"], 0.1]
                    ]
        else:
            self.allOperators = allOperators
        
        if not all(isinstance(elem, list) for elem in self.allOperators):
            raise ValueError("The first item of each element should be a list.")
        if not all(  all( isinstance(s, str)  for s in elem[0]) and   isinstance(elem[1], Real)    for elem in self.allOperators):
            raise ValueError("The first item of each element should be a list of strings and the second item a real number.")
        
    
    def get_lagrangian(self, max_terms = 50, allOperators = None):
        '''
        Construct the Lagrangian from a table of particle representations.

        Inputs:
        -------
        
        
            - model: int.
                This is the maximm number of terms allowed to have for the lagrangian. The lagrangian computation stops after this number is reached
        Outputs:
        ---------
        
            - lagrangian: sympy expression
                Sympy expression of Lagrangian.
            - parameters: list of sympy symbols
                List of symbols used in Lagrangian. Each element has four elements. The first element corresponds to the coefficient for that term. Note that this coefficient could have been set to one in the matrix multiplication since we can always factor out a coefficient in the mass matrix without loss of generality. The second element is the coefficient (usually of the form 10^{n} for some integer n) that denotes the supression from higher-order term. The third element is the sympy expressin of the lagrangian before evaluating any flavon vevs. The last entry is a list of the fields involve in this term writen as [PyDiscreteField, [Non-Abelian reps] ].
        '''
        fieldsForDiscrete = self.fields_for_discrete(self.model)
        self.particle_dict = self._get_particle_dict(fieldsForDiscrete)
        self._setallOperators(allOperators)


        
        lagrangian = 0
        parameters = []
        n_terms = 0
        
        # Loop through allPperators
        for i in range(len(self.allOperators)):
            if n_terms > max_terms:
                 break
            repeated_fields = self._label_repeated_fields(self.allOperators[i][0])
            list_fields = [self.particle_dict[s] for s in self.allOperators[i][0]]
            coeff = self.allOperators[i][1]
            lagrangian, parameters, n_terms = self._operatorToLag(list_fields, lagrangian, parameters, n_terms, coeff, max_terms, repeated_fields)


        return lagrangian, parameters

    
        

    def ParametersInAMatrix(self, mat):
        '''
        List the symbols of the parameters used in a matrix.

        Input:
            - mat: A sympy matrix.
                The matrix from which we will extract the parameter symbols.
        Output:
            - params_in_mat: List of symbols in the matrix.
                This are the sympy symbols which are the parameters in the matrix.
        '''
        # Extract all symbols in the matrix
        all_symbols = set()
        for element in mat:
            all_symbols.update(element.atoms(sp.Symbol))
        
        # Convert the set of symbols to a list
        list_symbols = list(all_symbols)
        
        # If the result is zero, return an empty list
        if list_symbols == [0]:
            list_symbols = []
        
        params_in_mat = sorted(list_symbols,key=lambda x: str(x))
        return params_in_mat
        
        
    def get_rank_mass_matrix(self,mass_matrix, parameters):
        """
        Helper function to calculate the rank of mass matrices using random numbers between 0.1 and 1 for the parameters.
        
        Inputs:
             - mass_matrix: sympy matrix
                   The mass matrix to calculate its ranks.
             - parameters: list paramerters obtained from get_lagrangian()
             
        Outsputs:
             - mass_rank: int
                   Rank of mass_matrix
        """
        
        all_params = [p[0] for p in parameters]
        test_replacement_list = {a: random.uniform(0.1, 1) for a in all_params}
        mass_matrix_test = np.array(mass_matrix.xreplace(self.parametersSet).xreplace(test_replacement_list).evalf().tolist(), dtype=complex)
        mass_rank = np.linalg.matrix_rank(mass_matrix_test)
        
        return mass_rank
        
    def count_eff_flavons(self, lagrangian):
        """
        Function to count flavons present in the lagrangian.
            
        Inputs:
        ------
             -lagrangian: sympy expression
                   This is obtained from get_lagrangian()
                   
        Outputs:
        -------
             - eff_flavons: int
                   Number of effective flavons present in the lagrangian
             - flavons_not_present: list of strings
                   List of strings with the flavons not present in the lagrangian
        
        
        """
        
        
        eff_flavons = 0
        flavons_Not_present = []
        
        for i in range(1, self.NumberOfTotalFlavons  + 1):
            if f"phi{i}" in str(lagrangian):
                eff_flavons = eff_flavons + 1
            else:
                flavons_Not_present.append(f"phi{i}")

        return eff_flavons, flavons_Not_present
        
        
        
    def _extract_symbols(self, obj, symbolToFind):
        
        """
        Helper fuction to extract symbols from give a symbol to find and flatten them in a list.
        Inputs:
        -------
            - obj: list of symbols
                    List to look for symbols that contain symbolToFind
            - symbolToFind: str
                    this is the symbol that will be found in the obj.
                    
        Outputs:
        --------
        
            - founds_symbols: list of symbols
                    List of symbols that contain symbolToFind
        
        """
            
            
        founds_symbols = []
        if isinstance(obj, Symbol):
            if symbolToFind in obj.name:
                founds_symbols.append(obj)
        elif isinstance(obj, (list, tuple)):
            for item in obj:
                founds_symbols.extend(self._extract_symbols(item, symbolToFind))
        return founds_symbols

        
    def get_mass_matrix(self, lagrangian ,parameters, left_fields, right_fields = None, higgs_fields = None,  factor_out = True):
    
        """
        This function obtains the sympy mass matrix for the lagrangian in terms of parameters. The mass matrix can be between a Dirac term in which case you need to specify the left_fields, the right_fields and the higgs_fields. You can also use it for a Majorana mass matrix, in which case you only specify left_fields.
        
        Inputs:
        -------
        
            - lagrangian: sympy symbol
                Lagrangian obtained from get_lagrangian.
            - parameters: list of sympy symbols
                Parameters obtained from get_lagrangian
            - left_fields: 2D list of strings
                Left fields that will be used for calculating the mass matrix
            - right_fields: 2D list of strings (optional, None by defaultu)
                Right fields that will be used for calculating the mass matrix
            - higgs_fields: 2D list of strings (optional, None by default)
                Higgs fields that will be used for calculating the mass matrix
            - factor_out: Boolean (True, by default)
                If true, it factors out one of the coefficients of the mass matrix. This can always be done for fitting observabbles. It makes the fit more efficiently
        
        
        Outputs:
        --------
        
            - mass_matrix: sympy matrix
                Dirac or Majorana mass matrix obtained from the lagrangian.
        
        """
        
        
        if left_fields not in self.particle_list_grouping:
            raise ValueError("left_fields must be in particle_list_grouping")

        if (right_fields is not None) and (right_fields not in self.particle_list_grouping):
            raise ValueError("right_fields must be in particle_list_grouping")

        if (higgs_fields is not None) and (higgs_fields not in self.particle_list_grouping):
            raise ValueError("higgs_fields must be in particle_list_grouping")
        
        
        minimum = {}
        left_fields_list = self.particle_dict[left_fields[0]]
        left_fields_flattened = self._extract_symbols(left_fields_list, left_fields[1])
        
        right_fields_list = self.particle_dict[right_fields[0]]
        right_fields_flattened = self._extract_symbols(right_fields_list, right_fields[1])
        
        if (higgs_fields is not None) and (right_fields is not None):
            higgs_fields_list = self.particle_dict[higgs_fields[0]]
            higgs_fields_flattened = self._extract_symbols(higgs_fields_list, higgs_fields[1])
            minimum.update({symbol: 1 for symbol in higgs_fields_flattened})
            self.parametersSet[higgs_fields_flattened[0]] = 1

        
        
        
        minimum.update({symbol: 0 for symbol in left_fields_flattened})
        minimum.update({symbol: 0 for symbol in right_fields_flattened})
        



        if lagrangian == 0:
            mass_matrix = sp.zeros(3, 3)
        else:
            mass_matrix = sp.Matrix(3, 3, lambda i, j: lagrangian.diff(left_fields_flattened[i], right_fields_flattened[j])).xreplace(minimum).xreplace(self.parametersSet)
            mass_matrix  = self.group._eliminate_small_real_imag(sp.simplify(mass_matrix))

        if factor_out:
            if len(self.ParametersInAMatrix(mass_matrix)) > 0:
                self.parametersSet[self.ParametersInAMatrix(mass_matrix)[0]] =  1


        mass_matrix = self.group._eliminate_small_real_imag(mass_matrix.xreplace(self.parametersSet))

        return mass_matrix


def make_latex_tableA4(file_path = None, model = None):


    """
    Reads a A4 model file and produce a LaTeX model table.

    Parameters
    ----------
    model_file : str
        Path to the file.
    model: dictionary
        Dictionary containing the file

    Returns
    -------
    str
        A LaTeX table environment as a string.
    """

    if file_path is not None:
        with open(file_path, "r") as file:
            data = json.load(file)

    if model is not None:
        data = model
    
    if file_path is None and model is None:
        raise ValueError("You need to give a model file or model")
        
    
    # Extract necessary components
    representations = data["representations"]
    particles = data["particles"]
    ZN_symmetry = data["ZN_symmetries"][0]
    fixed_flavon_vevs = data.get("fixed_flavon_vevs", [])
    
    # Format particle names with LaTeX subscripts
    def format_particle(name):
        if name == "Hu":
            return "H_u"
        elif name == "Hd":
            return "H_d"
        elif "phi" in name:
            name = name.replace("phi", "\\phi")
        for i, c in enumerate(name):
            if c.isdigit():
                return f"{name[:i]}_{name[i:]}"
        return f"{name}"

    # VEV mapping dictionary
    vev_map = {
        2: r"\left( \begin{array}{c} 1 \\ 0 \\ 0 \end{array} \right)",
        3: r"\left( \begin{array}{c} 1 \\ 1 \\ 1 \end{array} \right)",
        4: r"\left( \begin{array}{c} 0 \\ 1 \\ 0 \end{array} \right)",
        5: r"\left(\begin{array}{c} 0 \\ 0 \\ 1 \end{array} \right)",
        6: r"\left(\begin{array}{c} 0 \\ 1 \\ -1 \end{array}\right)",
        7: r"\left(\begin{array}{c} 1 \\ \omega^2 \\ \omega \end{array}\right)",
        8: r"\left(\begin{array}{c} 1 \\ \omega \\ \omega^2 \end{array}\right)",
     }

    # Build lists only for non-zero representation entries
    filtered_particles = []
    filtered_representations = []
    filtered_indices = []

    for idx, (p, r) in enumerate(zip(particles[:16], representations[:16])):
        if any(r):  # skip if all entries are zero
            filtered_particles.append(format_particle(p))
            filtered_representations.append(r)
            filtered_indices.append(idx)

    latex_lines = []

    # Header row (particle names)
    header_row = " & " + " & ".join(filtered_particles) + " \\\\"
    latex_lines.append("\\hline")
    latex_lines.append(header_row)
    latex_lines.append("\\hline")

    # Second row (A4 representations)
    a4_row = ["A_4"]
    for rep in filtered_representations:
        if rep[0] == 1:
            a4_row.append("\\mathbf{1}")
        elif rep[1] == 1:
            a4_row.append("\\mathbf{1'}")
        elif rep[2] == 1:
            a4_row.append("\\mathbf{1''}")
        elif rep[3] >= 1:
            a4_row.append("\\mathbf{3}")
        else:
            a4_row.append("")
    latex_lines.append(" & ".join(a4_row) + " \\\\")
    latex_lines.append("\\hline")

    # Third row (\mathbb{Z}_N charges)
    zn_row = [f"\\mathbb{{Z}}_{ZN_symmetry}"]
    for rep in filtered_representations:
        zn_row.append(str(rep[4] - 1))
    latex_lines.append(" & ".join(zn_row) + " \\\\")
    latex_lines.append("\\hline")


    # Fourth row (Vevs)
    vev_row = [r"\text{Vevs}"]
    for idx, name in zip(filtered_indices, particles[:16]):
        if "phi" in name and idx >= 10:
            vev_row.append(vev_map.get(fixed_flavon_vevs[idx-11], ""))
        else:
            vev_row.append("")
    latex_lines.append(" & ".join(vev_row) + " \\\\")
    latex_lines.append("\\hline")

    n_columns = len(filtered_particles) + 1
    latex_code = "\\begin{array}{|" + "|".join(["c"] * n_columns) + "|}\n" + "\n".join(latex_lines) + "\n\\end{array}"
    
    return latex_code
        
        
def make_latex_tableT19(file_path = None, model = None):
    """
    Reads a T19 model file and produce a LaTeX model table.

    Parameters
    ----------
    model_file : str
        Path to the file.
    model: dictionary
        Dictionary containing the file
        
    
    Returns
    -------
    str
        A LaTeX array environment as a string.
    """

    if file_path is not None:
        with open(file_path, "r") as file:
            data = json.load(file)

    if model is not None:
        data = model
    
    if file_path is None and model is None:
        raise ValueError("You need to give a model file or model")
        

    # Extract necessary components
    representations = data["representations"]
    particles = data["particles"]
    ZN_symmetry = data["ZN_symmetries"][0]
    fixed_flavon_vevs = data.get("fixed_flavon_vevs", [])

    # Format particle names with LaTeX subscripts
    def format_particle(name: str) -> str:
        if name == "Hu":
            return "H_u"
        elif name == "Hd":
            return "H_d"
        elif "phi" in name:
            name = name.replace("phi", "\\phi")
        for i, c in enumerate(name):
            if c.isdigit():
                return f"{name[:i]}_{name[i:]}"
        return f"{name}"

    # VEV mapping dictionary
    vev_map = {
        2: r"\left( \begin{array}{c} 1 \\ 0 \\ 0 \end{array} \right)",
        3: r"\left( \begin{array}{c} 1 \\ 1 \\ 1 \end{array} \right)",
        4: r"\left( \begin{array}{c} 0 \\ 1 \\ 0 \end{array} \right)",
        5: r"\left(\begin{array}{c} 0 \\ 0 \\ 1 \end{array} \right)",
        6: r"\left(\begin{array}{c} 0 \\ 1 \\ -1 \end{array}\right)",
        7: r"\left(\begin{array}{c} 1 \\ \omega^2 \\ \omega \end{array}\right)",
        8: r"\left(\begin{array}{c} 1 \\ \omega \\ \omega^2 \end{array}\right)",
    }

    # Build lists only for non-zero representation entries
    filtered_particles = []          # formatted names
    filtered_representations = []    # representation vectors
    filtered_indices = []            # original indices in 'particles'

    for idx, (p, r) in enumerate(zip(particles[:17], representations[:17])):
        if any(r):  # keep if any non-zero entry
            filtered_particles.append(format_particle(p))
            filtered_representations.append(r)
            filtered_indices.append(idx)

    latex_lines = []

    # Header row (particle names)
    header_row = " & " + " & ".join(filtered_particles) + " \\\\"
    latex_lines.append("\\hline")
    latex_lines.append(header_row)
    latex_lines.append("\\hline")

    # Second row (A4 representations)
    a4_row = ["T_{19}"]
    for rep in filtered_representations:
        if rep[0] == 1:
            a4_row.append("\\mathbf{1}")
        elif rep[1] == 1:
            a4_row.append("\\mathbf{1'}")
        elif rep[2] == 1:
            a4_row.append("\\mathbf{1''}")
        elif rep[3] >= 1:
            a4_row.append("\\mathbf{3_1}")
        elif rep[4] >= 1:
            a4_row.append("\\mathbf{\\overline{3}}_1")
        elif rep[5] >= 1:
            a4_row.append("\\mathbf{3_2}")
        elif rep[6] >= 1:
            a4_row.append("\\mathbf{3_3}")
        elif rep[7] >= 1:
            a4_row.append("\\mathbf{\\overline{3}}_2")
        elif rep[8] >= 1:
            a4_row.append("\\mathbf{\\overline{3}}_3")
        else:
            a4_row.append("")
    latex_lines.append(" & ".join(a4_row) + " \\\\")
    latex_lines.append("\\hline")

    # Third row (\mathbb{Z}_N charges)
    zn_row = [f"\\mathbb{{Z}}_{ZN_symmetry}"]
    for rep in filtered_representations:
        zn_row.append(str(rep[9] - 1))
    latex_lines.append(" & ".join(zn_row) + " \\\\")
    latex_lines.append("\\hline")

    # Fourth row (VEVs)
    vev_row = [r"\text{Vevs}"]
    # Maintain the same column order as filtered_particles by iterating filtered_indices
    for idx in filtered_indices:
        name_original = particles[idx]
        if "phi" in name_original and idx >= 10:
            j = idx - 11  # guard against OOB
            code = fixed_flavon_vevs[j] if 0 <= j < len(fixed_flavon_vevs) else None
            vev_row.append(vev_map.get(code, ""))
        else:
            vev_row.append("")
    latex_lines.append(" & ".join(vev_row) + " \\\\")
    latex_lines.append("\\hline")

    n_columns = len(filtered_particles) + 1
    latex_code = (
        "\\begin{array}{|"
        + "|".join(["c"] * n_columns)
        + "|}\n"
        + "\n".join(latex_lines)
        + "\n\\end{array}"
    )

    return latex_code


